// (c) 2017 Manfred Jeusfeld, license Creative Commons CC-BY 4.0
// compile Windows: javac -classpath c:\conceptbase\lib\classes\cb.jar LocalADOxxClient.java
// compile Linux: javac -classpath $CB_HOME/lib/classes/cb.jar LocalADOxxClient.java
// run Windows: java -classpath c:\conceptbase\lib\classes\cb.jar;. LocalADOxxClient
// run Linux: java -classpath javac -classpath $CB_HOME/lib/classes/cb.jar:. LocalADOxxClient

// Author: Manfred Jeusfeld
// Created: 2017-03-28 (2017-03-29)

// These classes are part of the cb.jar ConceptBase Java library, in particular LocalCBclient
// Like ADOxxClient but connects to a localhost (see sCBserver)
import i5.cb.api.*;


/**
ADOxxClient shows an example how a modeling tool like ADOxx can interact with a ConceptBase Server (CBserver).
The CBserver shall be preloaded with the ADOxx meta-2 model within the database module 'ADOxx'.
The CBShell script createADOCB.cbs on how to create such a database via 'cbshell createADOCB.cbs'.

To use this program successfully, you must make sure that a CBserver was started before on the 
Linux computer specified as sCBserver below. The following command on that computer could be used to
do so:
  cbserver -sm slave -r 1 -port 4002 -a yourusername -t low -d ADOCB -u nonpersistent &> log-4002.txt &
Note that it loads the database ADOCB that must have been created before on the same computer.
*/

public class LocalADOxxClient {

  private static LocalCBclient cbClient = null;
  public final static String sCBserver = "localhost";   // change this to the location of your CBserver
  public final static int iPort = 4002;                           // use port number 4002 to start the CBserver
  public final static String sTool = "ADOxxClient";
  public final static String sUser = "ADOxx";
  private static String answer;
  

// initConnection establishes the connection to the CBserver and switches to database module ADOxx

  public static void initConnection() {
      cbClient=new LocalCBclient();
      cbClient.connect(sCBserver,iPort,sTool,sUser);
      cbClient.cd("ADOxx");  // this is the module reserved for the tool ADOxx in the ConceptBase server

  }


// extractERMM is a stub function that generates the Telos frames for a given metamodel
// such as the ER metamodel.  Note that the frames are put into small pieces resembling
// triples. Instead of user-readable names like ENTITY, you may also use system-generated IDs from
// ADOxx, e.g. A1234. This does however not hold for the meta2 object like CLASS and ABSTRACT.
 
  public static String extractERMM() {
      return "EoR in ABSTRACT end\n "
            + "EoR in CLASS end\n"
            + "EoR with relationship hasAttribute: ATTRIBUTE end\n"
            + "ATTRIBUTE in CLASS end\n "
            + "ENTITY in CLASS end\n "
            + "ENTITY with inheritance super: EoR end\n"
            + "RELATION in CLASS end\n "
            + "RELATION with inheritance  super: EoR end\n"
            + "RELATION with relationship relates: EoR end\n ";
  }


// createMetaModel creates a meta-model in a sub-module of ADOxx. Such definitions are typically
// done by the ADOxx Development toolkit.

  public static void createMetaModel(String mmname, String telosframes) {
      cbClient.mkdir(mmname);  
      cbClient.cd(mmname);
      answer = cbClient.tells(telosframes);
      System.out.println("createMetaModel "+ mmname + ": " + answer);
  }



  public static String extractERMMQueries() {
      return "OBJECTwithSharedAttribute in QueryClass isA EoR with "
           + "  computed_attribute "
           + "    attr: ATTRIBUTE; "
           + "    other: EoR "
           + "  constraint "
           + "    c1: $ (this hasAttribute attr) and (other hasAttribute attr) and (this <> other) $ "
           + "end \n  "
           + "ENTITYwithoutAttribute in QueryClass isA EoR with "
           + "  constraint "
           + "    c1: $ (this in ENTITY) and not exists a/ATTRIBUTE (this hasAttribute a) $ "
           + "end\n ";
  }



 // defineQueries defines some analysis queries to check the integrity of an ER model.

  public static void defineMetaModelQueries(String mmname, String queries) {
      cbClient.cd(mmname);
      answer = cbClient.tells(queries);
      System.out.println("defineMetaModelQueries " + mmname + ": " + answer);
  }

 // extractLIT_MODEL is a stub function to extract an example ER model, here LIT_MODEL from
 // the modelling environment such as ADOxx Modelling Toolkit. 

  public static String extractLIT_MODEL() {
      return "ISBN in ATTRIBUTE end\n "
           + "Title in ATTRIBUTE end\n "
           + "Number in ATTRIBUTE end\n "
           + "Book in ENTITY end\n "
           + "Chapter in ENTITY end\n "
           + "isPartOf in RELATION end\n "
           + "Book with hasAttribute attr1: ISBN end\n "
           + "Book with hasAttribute attr2: Title end\n "
           + "Chapter with hasAttribute attr3: Number end\n "
           + "isPartOf with relates relates1: Book end\n "
           + "isPartOf with relates relates2: Chapter end\n "
           + "Employee in ENTITY end\n "
           + "Employee with hasAttribute attr4: Number end\n "
           + "Author in ENTITY end \n";
  }

 // createModel creates the telosframes into the database module mname, which itself
 // is a sub-module of the metamodel module mmname

  public static void createModel(String mmname, String mname, String telosframes) {
      cbClient.cd(mmname);
      cbClient.mkdir(mname);
      cbClient.cd(mname);
      answer = cbClient.tells(telosframes);
      System.out.println("createModel " + mmname + " " + mname + ": "   + answer);
  }


 // testFrame shows how integrity constraints work in ConceptBase. The class EoR of ERMM was
 // defined to be abstract, i.e. may have no explicit instances. The test object 'dummy'
 // is such an explicit instance. Trying to tell it, leads to an integrity violation.
 // and the rejection of the tell transaction.

  public static void testFrame(String mmname, String mname, String telosframe) {
     cbClient.cd(mmname+"-"+mname);
     answer = cbClient.tells(telosframe);
     if (!answer.equals("yes")) 
        System.out.println("testFrame ERROR: " + answer);
  }


  // main runs the example against the CBserver.   

  public static void main(String argv[])  {

      initConnection();

      createMetaModel( "ERMM",extractERMM() );

      defineMetaModelQueries( "ERMM",extractERMMQueries() );

      createModel("ERMM","LIT_MODEL",extractLIT_MODEL());

      testFrame("ERMM","LIT_MODEL","dummy in EoR end\n ");

      // analysis queries executed in database module ERMM-LIT_MODEL
      cbClient.cd("ERMM-LIT_MODEL");
      answer = cbClient.asks("OBJECTwithSharedAttribute", "FRAME");
      System.out.println("\nOBJECTwithSharedAttribute: \n" + answer);

      answer = cbClient.asks("ENTITYwithoutAttribute", "LABEL");
      System.out.println("\nENTITYwithoutAttribute: \n" + answer);

      answer = cbClient.disconnect();   // don't forget to disconnect
  }
}

